/**
* A class for a table having a one table row heading, adjustable row and column count and editable plain text entries
*/
function EditableTable() {
    this._buildDom();
}

/** HTMLTableElement instance acting as a UI */
EditableTable.prototype.node = null;

/** Number of columns */
Object.defineProperty(EditableTable.prototype, 'colCount', {
    enumerable: true,
    get() { return this.node.firstChild.firstChild.childNodes.length - 1;},
    set(value) {
        const headTr = this.node.firstChild.firstChild;
        const rows = [...this.node.childNodes[1].childNodes].slice(0, -1);
        if (!(value >= 0) || !Number.isFinite(value))
            return;
        while (value < this.colCount) {
            headTr.removeChild(headTr.childNodes[headTr.childNodes.length - 2]);
            rows.forEach(row => row.removeChild(row.childNodes[row.childNodes.length - 2]));
        }
        while (value > this.colCount) {
            const th = ScoTools.createHtml('th');
            ScoTools.createHtml('input', th, { size: 10 }).addEventListener('change', () => this._changed());
            ScoTools.createHtml('br', th);
            EditableTable._addBtn(th, '⬅️', 'Attribut nach links').addEventListener('click', () => { const pos = [...headTr.childNodes].indexOf(th); this.swapCols(pos, pos -1); });
            EditableTable._addBtn(th, '📋', 'Attribut duplizieren').addEventListener('click', () => { const pos = [...headTr.childNodes].indexOf(th); this.cloneCol(pos); });
            EditableTable._addBtn(th, '-', 'Attribut löschen').addEventListener('click', () => this.removeCol([...headTr.childNodes].indexOf(th)));
            EditableTable._addBtn(th, '➡️', 'Attribut nach rechts').addEventListener('click', () => { const pos = [...headTr.childNodes].indexOf(th); this.swapCols(pos, pos + 1); });
            headTr.insertBefore(th, headTr.lastChild);
            rows.forEach(row => {
                const td = ScoTools.createHtml('td');
                ScoTools.createHtml('input', td, { size: 10 }).addEventListener('change', () => this._changed());
                row.insertBefore(td, row.lastChild);
            });
        }
    }
});

/** Number of rows */
Object.defineProperty(EditableTable.prototype, 'rowCount', {
    enumerable: true,
    get() { return this.node.childNodes[1].childNodes.length - 1;},
    set(value) {
        const body = this.node.childNodes[1];
        if (!(value >= 0) || !Number.isFinite(value))
            return;
        while (value < this.rowCount)
            body.removeChild(body.childNodes[body.childNodes.length - 2]);
        while (value > this.rowCount) {
            const tr = ScoTools.createHtml('tr');
            for (let n = this.colCount; n; --n) {
                const td = ScoTools.createHtml('td', tr);
                ScoTools.createHtml('input', td, { size: 10 }).addEventListener('change', () => this._changed());
            }
            const lastTd = ScoTools.createHtml('td', tr);
            EditableTable._addBtn(lastTd, '⬆️', 'Datensatz hoch').addEventListener('click', () => { const pos = [...body.childNodes].indexOf(tr); this.swapRows(pos, pos -1); });
            EditableTable._addBtn(lastTd, '⬇️', 'Datensatz runter').addEventListener('click', () => { const pos = [...body.childNodes].indexOf(tr); this.swapRows(pos, pos +1); });
            EditableTable._addBtn(lastTd, '📋', 'Datensatz duplizieren').addEventListener('click', () => { const pos = [...body.childNodes].indexOf(tr); this.cloneRow(pos); });
            EditableTable._addBtn(lastTd, '-', 'Datensatz löschen').addEventListener('click', () => this.removeRow([...body.childNodes].indexOf(tr)));
            body.insertBefore(tr, body.lastChild);
        }
    }
});

/**
* Add a button with remove or add icon to DOM
* @return HTMLButtonElement     Added button
*/
EditableTable._addBtn = function(
    parentNode, // DOM node to be parent element of button
    type,       // Type of button: '+', '-', '⬅️', '➡️', '⬆️', '⬇️' or '📋'
    title,      // Hint to show
    desc        // Optional: Description to use
) {
    let text;
    switch (type) {
        case '📋':
        case '⬅️':
        case '➡️':
        case '⬆️':
        case '⬇️':
            text = type;
            break;
        case '-':
            text = '❌';
            break;
        case '+':
        default:
            text = '➕';
    }
    if (desc)
        text += ' ' + desc;
    const btn = ScoTools.createHtml('button', parentNode, { title }, text);
    return btn;
}

/**
* Create DOM representation
* @return undefined
*/
EditableTable.prototype._buildDom = function() {
    this.node = ScoTools.createHtml('table', null, { 'class': 'editable' });
    const head = ScoTools.createHtml('thead', this.node);
    const headTr = ScoTools.createHtml('tr', head);
    const headLastCell = ScoTools.createHtml('th', headTr);
    EditableTable._addBtn(headLastCell, '+', 'Attribut ergänzen').addEventListener('click', () => this.addColumn());
    const body = ScoTools.createHtml('tbody', this.node);
    const bodyTr = ScoTools.createHtml('tr', body);
    const bodyLastCell = ScoTools.createHtml('td', bodyTr);
    EditableTable._addBtn(bodyLastCell, '+', 'Datensatz ergänzen', 'Datensatz').addEventListener('click', () => this.addRow());
}

/** Function to call whenever table changed */
EditableTable.prototype.onchange = function() {}

/**
* Notify master on content change
* @return undefined
*/
EditableTable.prototype._changed = function() {
    if ((typeof this.onchange == 'function') || (this.onchange instanceof Function))
        this.onchange();
}

/**
* Add another column and focus column title input
* @return undefined
*/
EditableTable.prototype.addColumn = function() {
    ++this.colCount;
    const addedTh = this.node.firstChild.getElementsByTagName('th')[this.colCount - 1];
    addedTh.firstChild.focus();
    this._changed();
}

/**
* Add another row and focus its first input
* @return undefined
*/
EditableTable.prototype.addRow = function() {
    ++this.rowCount;
    const addedTr = this.node.childNodes[1].getElementsByTagName('tr')[this.rowCount - 1];
    addedTr.firstChild.firstChild.focus();
    this._changed();
}

/**
* Remove row and focus first next input
* @return undefined
*/
EditableTable.prototype.removeRow = function(
    row     // Number of row (starting at 0) to remove
) {
    const body = this.node.childNodes[1];
    const rows = body.childNodes;
    row = Math.max(0, Math.min(rows.length - 2, Number.isFinite(row) ? row : Infinity));
    body.removeChild(rows[row]);
    rows[row].firstChild.firstChild.focus();
    this._changed();
}

/**
* Remove column and focus next column's input
* @return undefined
*/
EditableTable.prototype.removeCol = function(
    col     // Number of column to remove (starting at 0)
) {
    const headTr = this.node.firstChild.firstChild;
    const rows = [...this.node.childNodes[1].childNodes].slice(0, -1).concat(headTr);
    col = Math.max(0, Math.min(headTr.childNodes.length - 2, Number.isFinite(col) ? col : Infinity));
    rows.forEach(row => row.removeChild(row.childNodes[col]));
    headTr.childNodes[col].firstChild.focus();
    this._changed();
}

/**
* Clone column
* @return undefined
*/
EditableTable.prototype.cloneCol = function(
    col     // Number of column to clone (starting at 0)
) {
    if ((col < 0) || (col >= this.colCount))
        return;
    const newIndex = this.colCount;
    ++this.colCount;
    [...this.node.getElementsByTagName('tr')].forEach(tr => {
        const inputs = tr.getElementsByTagName('input');
        if (inputs.length >= newIndex)
            inputs[newIndex].value = inputs[col].value;
    });
    this.node.firstChild.firstChild.lastChild.previousSibling.firstChild.focus();
    this._changed();
}

/**
* Clone row
* @return undefined
*/
EditableTable.prototype.cloneRow = function(
    row     // Number of row to clone (starting at 0)
) {
    if ((row < 0) || (row >= this.rowCount))
        return;
    const newIndex = this.rowCount;
    ++this.rowCount;
    const rows = this.node.lastChild.getElementsByTagName('tr');
    const source = [...rows[row].getElementsByTagName('input')].map(input => input.value);
    [...rows[newIndex].getElementsByTagName('input')].map((input, index) => (input.value = source[index]));
    rows[newIndex].lastChild.previousSibling.firstChild.focus();
    this._changed();
}

/**
* Swaps content of two columns
* @return undefined
*/
EditableTable.prototype.swapCols = function(
    index1,
    index2
) {
    const min = Math.min(index1, index2);
    const max = Math.max(index1, index2);
    if ((min < 0) || (min == max) || (max >= this.colCount))
        return;
    [...this.node.getElementsByTagName('tr')].forEach(tr => {
        const inputs = tr.getElementsByTagName('input');
        if (inputs.length >= max)
            [inputs[min].value, inputs[max].value] = [inputs[max].value, inputs[min].value];
    });
    this._changed();
}

/**
* Swaps content of two rows
* @return undefined
*/
EditableTable.prototype.swapRows = function(
    index1,
    index2
) {
    const min = Math.min(index1, index2);
    const max = Math.max(index1, index2);
    if ((min < 0) || (min == max) || (max >= this.rowCount))
        return;
    const rows = this.node.lastChild.getElementsByTagName('tr');
    const row1 = [...rows[min].getElementsByTagName('input')];
    const row2 = [...rows[max].getElementsByTagName('input')];
    row1.forEach((input1, index) => ([input1.value, row2[index].value] = [row2[index].value, input1.value]));
    this._changed();
}

/**
* Get all inputed values
* @return Array     Two dimensional array of all values
*/
EditableTable.prototype.getValues = function() {
    return [...this.node.getElementsByTagName('tr')].slice(0, -1).map(tr => [...tr.getElementsByTagName('input')].map(input => input.value));
}

/**
* Convert content to CSV string
* @return String        CSV string representing table's content
*/
EditableTable.prototype.toString = function() {
    const BACKSLASH_REGEX = new RegExp("\\\\", "g");
    const SEMICOLON_REGEX = new RegExp(";", "g");
    return [...this.node.getElementsByTagName('tr')].slice(0, -1).map(tr => [...tr.getElementsByTagName('input')].map(input => input.value.replace(BACKSLASH_REGEX, '\\\\').replace(SEMICOLON_REGEX, '\\;')).join(';')).join('\n');
}

/**
* Convert CSV string to two dimensional array
* @return Array     Data represented by string
*/
EditableTable.stringToArray = function(
    str         // CSV string (values seperated by ; and new lines)
) {
    return data = str.split('\n').map(row => {
        let result = [''];
        let escapeNextChar = false;
        for (let i = 0; i < row.length; i++) {
            if (escapeNextChar) {
                result[result.length - 1] += row[i];
                escapeNextChar = false;
                continue;
            }
            switch (row[i]) {
                case '\\':
                    escapeNextChar = true;
                    break;
                case ';':
                    result.push('');
                    break;
                default:
                    result[result.length - 1] += row[i];
            }
        }
        return result;
    });
}

/**
* Load CSV string
* @return undefined
*/
EditableTable.prototype.fromString = function(
    str         // CSV string (values seperated by ; and new lines)
) {
    const data = EditableTable.stringToArray(str);
    this.rowCount = data.length - 1;
    this.colCount = data.reduce((prev, curr) => Math.max(prev, curr.length), 0);
    [...this.node.getElementsByTagName('tr')].forEach((tr, y) => [...tr.getElementsByTagName('input')].forEach((input, x) =>
        input.value = (data[y][x] === undefined) ? '' : data[y][x]
    ));
    this._changed();
}
